#!/usr/bin/env node

/*
 * BSD 3-Clause License
 *
 * Copyright (c) 2015, Nicolas Riesco and others as credited in the AUTHORS file
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

var assert = require("assert");
var fs = require("fs");

var MessagingTestEngine = require("./mte");

describe("A Kernel instance", function() {
    describe("using protocol v4.1", function() {
        var mte = new MessagingTestEngine("4.1");

        before(function(done) {
            mte.init(done);
        });

        after(function() {
            mte.dispose();
        });

        testHeartBeat(mte);
        testKernelInfoRequest(mte);
        testInputRequest(mte);
        testMessagingProtocol(mte);
        testHeartBeat(mte);
    });

    describe("using protocol v5.0", function() {
        var mte = new MessagingTestEngine("5.0");

        before(function(done) {
            mte.init(done);
        });

        after(function() {
            mte.dispose();
        });

        testHeartBeat(mte);
        testKernelInfoRequest(mte);
        testMessagingProtocol(mte);
        testInputRequest(mte);
        testHeartBeat(mte);
    });
});

function testHeartBeat(mte) {
    it("has heart beats", function(done) {
        var hbCount = 10;

        mte.socket.hb.on("message", onHeartBeat);

        mte.socket.hb.send();

        function onHeartBeat() {
            hbCount--;

            if (hbCount > 0) {
                mte.socket.hb.send();
            } else {
                mte.socket.hb.removeListener("message", arguments.callee);
                done();
            }
        }
    });
}

function testKernelInfoRequest(mte) {
    var description = "kernel_info_request (content test)";

    it("replies correctly to " + description, function(done) {
        var request = {
            shell: {
                header: {
                    msg_type: "kernel_info_request",
                    version: mte.version.protocol,
                }
            }
        };

        var responses = testKernelInfoReply;

        var testCase = {
            description: description,
            request: request,
            responses: responses,
            done: done,
        };
        mte.run(testCase);
    });

    function testKernelInfoReply(response, socketName) {
        if (socketName !== "shell") {
            return true;
        }

        if (response.header.msg_type !== "kernel_info_reply") {
            return true;
        }

        assert.deepEqual(
            response.metadata, {},
            "Error: no metadata"
        );

        assert.deepEqual(
            response.content, mte.kernel.kernelInfoReply,
            "Error: bad content"
        );

        return false;
    }
}

function testMessagingProtocol(mte) {
    var testCases = JSON.parse(fs.readFileSync(mte.path.testMessages));

    var major = parseInt(mte.version.protocol.split(".", 1)[0]);
    if (major < 5) {
        testCases = testCases.concat(JSON.parse(
            fs.readFileSync(mte.path.testMessagesV4)
        ));
    } else {
        testCases = testCases.concat(JSON.parse(
            fs.readFileSync(mte.path.testMessagesV5)
        ));
    }

    var wait = 0;
    testCases.forEach(function(testCase) {
        it("replies correctly to " + testCase.description, function(done) {
            testCase.done = done;

            wait += 0;
            setTimeout(function() {
                mte.run(testCase);
            }, wait);
        });
    });
}

function testInputRequest(mte) {
    it("can run $$.input() and receive a response", function(done) {
        var testCase = {
            description: "$$.input()",
            request: {
                "shell": {
                    "header": {
                        "msg_type": "execute_request"
                    },
                    "content": {
                        "allow_stdin": true,
                        "code": "$$.input({prompt:'?', password: true}, function(error, reply) {$$.done(reply)});" // eslint-disable-line max-len
                    }
                }
            },
            responses: function testInputReply(response, socketName) {
                if (socketName === "stdin") {
                    response.respond(
                        mte.socket.stdin,
                        "input_reply", {
                            value: "opensesame"
                        }
                    );
                    return true;
                } else if (socketName !== "iopub") {
                    return true;
                }

                var major = parseInt(mte.version.protocol.split(".", 1)[0]);
                var msg_type = (major < 5) ? "pyout" : "execute_result";
                if (response.header.msg_type !== msg_type) {
                    return true;
                }

                assert.deepEqual(
                    response.content.data, {
                        "text/plain": "'opensesame'"
                    },
                    "Error: bad content"
                );

                return false;
            },
            done: done,
        };

        mte.run(testCase);
    });
}
